@@ -18,6 +18,8 @@ import math
 import os
 import sys
 import threading
+from datetime import datetime
+from typing import Dict, Union, Optional, List
 
 import requests as rq
 from joblib import Parallel, delayed
@@ -27,6 +29,7 @@ from tqdm import tqdm
 from urllib3 import Retry
 
 from .http_token_authentication import HTTPTokenAuth
+from .query import Condition, Conditions, Equal, GreaterThan, LessThan, Or
 
 logger = logging.getLogger(__name__)
 
@@ -46,15 +49,17 @@ class DLDSClient:
 
     """
 
-    def __init__(self, username=None, password=None, auth_token=None, http_retries=10, parallel_requests=16,
-                 server_url='https://api.vision.data-spree.com/api', verify_ssl=True, verify_s3_ssl=True):
+    def __init__(self, username: Optional[str] = None, password: Optional[str] = None, auth_token: Optional[str] = None,
+                 http_retries: int = 10, parallel_requests: int = 16,
+                 server_url: str = 'https://api.vision.data-spree.com/api', verify_ssl: bool = True,
+                 verify_s3_ssl: bool = True):
 
         self.username = username
         self.password = password
         self.auth_token = auth_token
 
         self.auth = None
-        if auth_token is not None:
+        if auth_token:
             self.auth = HTTPTokenAuth(auth_token)
         elif username is not None and password is not None:
             self.auth = HTTPBasicAuth(username, password)
@@ -89,6 +94,36 @@ class DLDSClient:
                              self.http.verify,
                              self.s3_http.verify,))
 
+    @staticmethod
+    def get_dlds_dataset_item_date_format_string() -> str:
+        return '%Y-%m-%dT%H:%M:%S.%fZ'
+
+    @staticmethod
+    def get_dlds_dataset_item_date_format_string_short() -> str:
+        return '%Y-%m-%dT%H:%M:%SZ'
+
+    @staticmethod
+    def parse_dlds_dataset_item_date(date_string: Optional[str]) -> Optional[datetime]:
+        if date_string is None:
+            return None
+
+        try:
+            return datetime.fromisoformat(date_string)
+        except (TypeError, ValueError):
+            pass
+
+        try:
+            return datetime.strptime(date_string, DLDSClient.get_dlds_dataset_item_date_format_string())
+        except (TypeError, ValueError):
+            pass
+
+        try:
+            return datetime.strptime(date_string, DLDSClient.get_dlds_dataset_item_date_format_string_short())
+        except (TypeError, ValueError):
+            pass
+
+        return None
+
     @staticmethod
     def dlds_dir():
         base_dir = os.path.expanduser('~')
@@ -102,6 +137,208 @@ class DLDSClient:
             pass
         return dlds_dir
 
+    def create_dataset(self,
+                       dataset_name,
+                       classification_class_label_ids=None,
+                       key_point_ids=None,
+                       roi_classification_class_label_ids=None,
+                       roi_key_point_ids=None,
+                       object_detection_class_label_ids=None,
+                       object_detection_key_point_ids=None,
+                       object_detection_masks=False,
+                       semantic_segmentation_class_label_ids=None):
+        """Create a new dataset.
+
+        :param dataset_name: Name of the new dataset.
+        :param classification_class_label_ids: IDs of the class labels for the classification of the whole image.
+        :param key_point_ids: IDs of the key points that can be annotated for the whole image.
+        :param roi_classification_class_label_ids: IDs of the class label for the classification of regions of interest.
+        :param roi_key_point_ids: IDs of the key points that can be annotated for each region of interest.
+        :param object_detection_class_label_ids: IDs of the class labels for object annotations.
+        :param object_detection_key_point_ids: IDs of the key points that can be annotated for individual objects.
+        :param object_detection_masks: Set to True if it should be possible to annotate individual objects with segmentation masks.
+        :param semantic_segmentation_class_label_ids: IDs of the class labels for annotating each pixel with a class label (semantic segmentation of the image).
+        :return: ID of the newly created dataset or -1 if dataset creation fails.
+        """
+
+        dataset = {
+            'name': dataset_name,
+            'classification': classification_class_label_ids is not None,
+            'key_points': key_point_ids is not None,
+            'roi_classification': roi_classification_class_label_ids is not None,
+            'roi_key_points': roi_key_point_ids is not None,
+            'object_detection': object_detection_class_label_ids is not None,
+            'object_detection_key_points': object_detection_key_point_ids is not None,
+            'object_detection_mask': object_detection_masks,
+            'semantic_segmentation': semantic_segmentation_class_label_ids is not None,
+        }
+
+        if dataset['classification']:
+            dataset['classification_class_label_ids'] = classification_class_label_ids
+
+        if dataset['key_points']:
+            dataset['key_point_ids'] = key_point_ids
+
+        if dataset['roi_classification']:
+            dataset['roi_classification_class_label_ids'] = roi_classification_class_label_ids
+
+        if dataset['roi_key_points']:
+            dataset['roi_key_point_ids'] = roi_key_point_ids
+
+        if dataset['object_detection']:
+            dataset['object_detection_class_label_ids'] = object_detection_class_label_ids,
+
+        if dataset['object_detection_key_points']:
+            dataset['object_detection_key_point_ids'] = object_detection_key_point_ids
+
+        if dataset['semantic_segmentation']:
+            dataset['semantic_segmentation_class_label_ids'] = semantic_segmentation_class_label_ids
+
+        try:
+            response = self.http.post(self.server_url + '/datasets/', data=dataset)
+            response.raise_for_status()
+            data = response.json()
+            return data['id']
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return -1
+
+    def get_dataset(self, dataset_id: int):
+        """Get information about a dataset.
+
+        Example:
+            >>> {
+            >>>    'id': 46,
+            >>>    'total_items': 104,
+            >>>    'uploaded_items': 0,
+            >>>    'annotated_items': 0,
+            >>>    'reviewed_items': 104,
+            >>>    'ignored_items': 0,
+            >>>    'created_by': {'id': 1, 'username': 'admin'},
+            >>>    'name': 'Traffic II',
+            >>>    'created_date': '2019-07-29T10:06:22.454871Z',
+            >>>    'updated_date': '2019-07-29T10:06:22.454914Z',
+            >>>    'classification': False,
+            >>>    'key_points': False,
+            >>>    'roi_classification': False,
+            >>>    'roi_key_points': False,
+            >>>    'object_detection': True,
+            >>>    'object_detection_key_points': False,
+            >>>    'object_detection_mask': False,
+            >>>    'semantic_segmentation': False,
+            >>>    'classification_class_labels': [],
+            >>>    'key_point_names': [],
+            >>>    'roi_classification_class_labels': [],
+            >>>    'roi_key_points_names': [],
+            >>>    'object_detection_class_labels': [{'id': 4, 'name': 'car'},
+            >>>     {'id': 6, 'name': 'truck'},
+            >>>     {'id': 7, 'name': 'pedestrian'},
+            >>>     {'id': 9, 'name': 'cyclist'},
+            >>>     {'id': 41, 'name': 'traffic light'},
+            >>>     {'id': 119, 'name': 'traffic sign'}],
+            >>>    'object_detection_key_point_names': [],
+            >>>    'semantic_segmentation_class_labels': []
+            >>> }
+
+        :param dataset_id: ID of the dataset to receive information.
+        :return: Dictionary containing information of the requested datatset.
+        """
+        try:
+            response = self.http.get(self.server_url + '/datasets/{}/'.format(dataset_id))
+            if response.status_code == 200:
+                dataset = response.json()
+                return dataset
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return None
+
+    def get_dataset_items(self, dataset_id: int,
+                          class_id: Optional[int],
+                          created_date_since: Optional[Union[str, datetime]] = None,
+                          created_date_until: Optional[Union[str, datetime]] = None,
+                          status: Optional[List[str]] = None,
+                          query_expression: Optional[Condition] = None,
+                          free_query_str: Optional[str] = None) -> Optional[List[Dict]]:
+        """
+        Return all dataset items that match the filter compiled from this functions' parameters.
+
+        :param dataset_id:          Identifier of the dataset whose content be downloaded.
+        :param class_id:            Set this to only download items of the specified class_id
+        :param created_date_since:  Set this to filter out dataset items created at created_date_since or earlier.
+        :param created_date_until:  Set this to filter out dataset items created at created_date_until or later.
+        :param status:              If not set, return the dataset items irrespective of their status.
+                                    Otherwise, add an OR filter to the query on the content of the status list.
+        :param query_expression:    Set this field to supply a custom query expression to the query.
+        :param free_query_str:      Set this field to supply a custom query expression as a string to the query.
+        :return:                    A list of dataset items or None, if a problem occurred.
+        """
+        dataset = self.get_dataset(dataset_id)
+        if dataset is None:
+            return None
+
+        items_per_request: int = 5000
+
+        def format_time(t: Optional[Union[str, datetime]]) -> Optional[str]:
+            return t if t is None or type(t) == str else t.isoformat()
+
+        # Compile conditions
+        condition_str: str = Conditions.create(Equal(var='class', val=class_id),
+                                               Equal(var='dataset__id', val=dataset_id),
+                                               Equal(var='page_size', val=items_per_request),
+                                               GreaterThan(var='created_date', val=format_time(created_date_since)),
+                                               LessThan(var='created_date', val=format_time(created_date_until)),
+                                               Or(*(Equal(var='status', val=s) for s in status)) if status else None,
+                                               query_expression, exclude_empty=True, do_url_escape=True)
+
+        if free_query_str is not None:
+            if not free_query_str.startswith('&'):
+                free_query_str = f'&{free_query_str}'
+            condition_str += free_query_str
+
+        # Compile URL
+        url: Optional[str] = f'{self.server_url}/dataset-items/?{condition_str}'
+
+        all_items: List[Dict] = list()
+        while url is not None:
+
+            logger.debug(f'Get dataset items from "{url}"')
+            try:
+                response: rq.models.Response = self.http.get(url)
+                response.raise_for_status()
+                response_json = response.json()
+
+                items: Optional[List[Dict]] = response_json.get('results')
+                url = response_json.get('next')
+
+                if not items:
+                    break
+
+                all_items.extend(items)
+
+            except rq.exceptions.RequestException as e:
+                logger.error(e)
+                return None
+
+        return all_items
+
+    def delete_dataset(self, dataset_id):
+        """Delete a dataset.
+
+        :param dataset_id: id of the dataset
+        :return: 'True' if success else 'False'
+        """
+
+        try:
+            response = self.http.delete(f'{self.server_url}/datasets/{dataset_id}')
+            response.raise_for_status()
+            return True
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return False
+
     def download_dataset(self, directory, dataset_id, n_items=-1, accepted_status=None):
         """Download a dataset from data spree vision platform.
 
@@ -170,79 +407,28 @@ class DLDSClient:
 
         logger.info('Download completed. (dataset ID: {})'.format(dataset_id))
 
-    def download_dataset_item(self, item_id, image_dir=None, annotation_dir=None, accepted_status=None,
-                              return_item=False):
-        """Download a dataset item into the specified directories (image and annotations).
+    def create_data_subset(self, subset_name, dataset_id, description=''):
+        """Add a subset to an existing dataset.
 
-        :param item: Dictionary containing the item details (e.g., ID, name, status).
-        :param dataset_id: ID of the dataset that the item belongs to.
-        :param image_dir: Output directory for the image.
-        :param annotation_dir: Output directory for the annotations.
-        :param accepted_status: List of accepted status. If the item status is not in this list, it will not be downloaded.
-        :return:
+        :param subset_name: name of the subset
+        :param dataset_id: id of the parent dataset
+        :param description: string description
+        :return: newly created subset or 'None' if subset creation fails.
         """
-        if image_dir == None and annotation_dir == None and not return_item:
-            raise ValueError('arguments not correct')
 
-        # load item details
+        subset_dict = {
+            'name': subset_name,
+            'description': description,
+            'dataset': dataset_id,
+        }
+
         try:
-            response = self.http.get(self.server_url + '/dataset-items/{}'.format(item_id))
+            response = self.http.post(f'{self.server_url}/data-subsets/', data=subset_dict)
             response.raise_for_status()
+            data = response.json()
+            return data
         except rq.exceptions.RequestException as e:
             logger.error(e)
-            return None
-
-        item_details = json.loads(response.content)
-        if accepted_status is not None:
-            if item_details['status'] not in accepted_status:
-                return None
-        image_url = item_details['data']
-
-        # download image
-        try:
-            response = self.s3_http.get(image_url, stream=True)
-            response.raise_for_status()
-        except rq.exceptions.RequestException as e:
-            logger.error(e)
-
-        item_id = item_details['id']
-        dataset_id = item_details['dataset_id']
-        item_name = item_details['name'].split('/')[-1]
-        item_name = '{}_{}_{}'.format(dataset_id, item_id, item_name)
-
-        image_bytes = None
-        if image_dir:
-            try:
-                os.makedirs(image_dir)
-            except FileExistsError:
-                pass
-            image_file_path = os.path.join(image_dir, item_name)
-
-            if return_item:
-                image_bytes = response.content
-
-            with open(image_file_path, 'wb') as f:
-                if image_bytes is None:
-                    for chunk in response.iter_content(4096):
-                        f.write(chunk)
-                else:
-                    f.write(image_bytes)
-
-        if annotation_dir:
-            try:
-                os.makedirs(annotation_dir)
-            except FileExistsError:
-                pass
-            # write annotations to file
-            annotations = item_details['annotations']
-            annotation_file_name = os.path.splitext(item_name)[0] + '.json'
-            annotation_file_path = os.path.join(annotation_dir, annotation_file_name)
-            with open(annotation_file_path, 'w') as f:
-                json.dump(annotations, f, indent=2)
-
-        if return_item:
-            item_details['image'] = image_bytes
-            return item_details
 
         return None
 
@@ -271,123 +457,6 @@ class DLDSClient:
 
         return None
 
-    def get_dataset(self, dataset_id):
-        """Get information about a dataset.
-
-        Example:
-            >>> {
-            >>>    'id': 46,
-            >>>    'total_items': 104,
-            >>>    'uploaded_items': 0,
-            >>>    'annotated_items': 0,
-            >>>    'reviewed_items': 104,
-            >>>    'ignored_items': 0,
-            >>>    'created_by': {'id': 1, 'username': 'admin'},
-            >>>    'name': 'Traffic II',
-            >>>    'created_date': '2019-07-29T10:06:22.454871Z',
-            >>>    'updated_date': '2019-07-29T10:06:22.454914Z',
-            >>>    'classification': False,
-            >>>    'key_points': False,
-            >>>    'roi_classification': False,
-            >>>    'roi_key_points': False,
-            >>>    'object_detection': True,
-            >>>    'object_detection_key_points': False,
-            >>>    'object_detection_mask': False,
-            >>>    'semantic_segmentation': False,
-            >>>    'classification_class_labels': [],
-            >>>    'key_point_names': [],
-            >>>    'roi_classification_class_labels': [],
-            >>>    'roi_key_points_names': [],
-            >>>    'object_detection_class_labels': [{'id': 4, 'name': 'car'},
-            >>>     {'id': 6, 'name': 'truck'},
-            >>>     {'id': 7, 'name': 'pedestrian'},
-            >>>     {'id': 9, 'name': 'cyclist'},
-            >>>     {'id': 41, 'name': 'traffic light'},
-            >>>     {'id': 119, 'name': 'traffic sign'}],
-            >>>    'object_detection_key_point_names': [],
-            >>>    'semantic_segmentation_class_labels': []
-            >>> }
-
-        :param dataset_id: ID of the dataset to receive information.
-        :return: Dictionary containing information of the requested datatset.
-        """
-        try:
-            response = self.http.get(self.server_url + '/datasets/{}/'.format(dataset_id))
-            if response.status_code == 200:
-                dataset = response.json()
-                return dataset
-        except rq.exceptions.RequestException as e:
-            logger.error(e)
-
-        return None
-
-    def create_dataset(self,
-                       dataset_name,
-                       classification_class_label_ids=None,
-                       key_point_ids=None,
-                       roi_classification_class_label_ids=None,
-                       roi_key_point_ids=None,
-                       object_detection_class_label_ids=None,
-                       object_detection_key_point_ids=None,
-                       object_detection_masks=False,
-                       semantic_segmentation_class_label_ids=None):
-        """Create a new dataset.
-
-        :param dataset_name: Name of the new dataset.
-        :param classification_class_label_ids: IDs of the class labels for the classification of the whole image.
-        :param key_point_ids: IDs of the key points that can be annotated for the whole image.
-        :param roi_classification_class_label_ids: IDs of the class label for the classification of regions of interest.
-        :param roi_key_point_ids: IDs of the key points that can be annotated for each region of interest.
-        :param object_detection_class_label_ids: IDs of the class labels for object annotations.
-        :param object_detection_key_point_ids: IDs of the key points that can be annotated for individual objects.
-        :param object_detection_masks: Set to True if it should be possible to annotate individual objects with segmentation masks.
-        :param semantic_segmentation_class_label_ids: IDs of the class labels for annotating each pixel with a class label (semantic segmentation of the image).
-        :return: ID of the newly created dataset or -1 if dataset creation fails.
-        """
-
-        dataset = {
-            'name': dataset_name,
-            'classification': classification_class_label_ids is not None,
-            'key_points': key_point_ids is not None,
-            'roi_classification': roi_classification_class_label_ids is not None,
-            'roi_key_points': roi_key_point_ids is not None,
-            'object_detection': object_detection_class_label_ids is not None,
-            'object_detection_key_points': object_detection_key_point_ids is not None,
-            'object_detection_mask': object_detection_masks,
-            'semantic_segmentation': semantic_segmentation_class_label_ids is not None,
-        }
-
-        if dataset['classification']:
-            dataset['classification_class_label_ids'] = classification_class_label_ids
-
-        if dataset['key_points']:
-            dataset['key_point_ids'] = key_point_ids
-
-        if dataset['roi_classification']:
-            dataset['roi_classification_class_label_ids'] = roi_classification_class_label_ids
-
-        if dataset['roi_key_points']:
-            dataset['roi_key_point_ids'] = roi_key_point_ids
-
-        if dataset['object_detection']:
-            dataset['object_detection_class_label_ids'] = object_detection_class_label_ids,
-
-        if dataset['object_detection_key_points']:
-            dataset['object_detection_key_point_ids'] = object_detection_key_point_ids
-
-        if dataset['semantic_segmentation']:
-            dataset['semantic_segmentation_class_label_ids'] = semantic_segmentation_class_label_ids
-
-        try:
-            response = self.http.post(self.server_url + '/datasets/', data=dataset)
-            response.raise_for_status()
-            data = response.json()
-            return data['id']
-        except rq.exceptions.RequestException as e:
-            logger.error(e)
-
-        return -1
-
     def get_data_subset_items(self, subset_id):
         """Get all subset items for the subset given by its ID.
 
@@ -420,60 +489,55 @@ class DLDSClient:
 
         return all_items
 
-    def get_dataset_items(self, dataset_id):
-        """Get all dataset items for the dataset given by its ID.
+    def delete_data_subset(self, subset_id):
+        """Delete a subset to an existing dataset.
 
-        :param dataset_id: Items of the provided dataset ID will be downloaded.
-        :return: List of items or None in case of errors.
+        :param subset_id: id of the subset
+        :return: 'True' if success else 'False'
         """
 
-        dataset = self.get_dataset(dataset_id)
-        if dataset is None:
-            return None
+        try:
+            response = self.http.delete(f'{self.server_url}/data-subsets/{subset_id}')
+            response.raise_for_status()
+            return True
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
 
-        n_items = dataset['total_items']
-        items_per_request = 5000
-        n_requests = int(math.ceil(n_items / items_per_request))
-
-        all_items = []
-        for i in range(n_requests):
-            offset = i * items_per_request
-            url = self.server_url + '/datasets/{}/items/?limit={}&offset={}'.format(dataset_id, items_per_request,
-                                                                                    offset)
-            logger.debug('Get dataset items from "{}"'.format(url))
-            try:
-                response = self.http.get(url)
-                response.raise_for_status()
-                items = response.json()['results']
-                all_items.extend(items)
-            except rq.exceptions.RequestException as e:
-                logger.error(e)
-                return None
-
-        return all_items
+        return False
 
     def create_dataset_item(self,
-                            image_file_path,
+                            image_file,
                             dataset_id,
-                            annotations={},
+                            annotations=None,
                             image_file_name=None,
-                            status=None):
+                            status=None,
+                            created_date=None):
         """Create a new dataset item including the image and the associated annotations.
 
-        :param image_file_path: File path to the image that belongs to the new item. This image will be uploaded.
+        :param image_file: File path to the image that belongs to the new item. This image will be uploaded or image bytes to upload
         :param dataset_id: The new item will be created for the provided dataset ID.
         :param annotations: Dictionary of the annotations
-        :param image_file_name: Name of the image that should be uploaded. Set to `None` to use the file name derived from `image_file_path`.
+        :param image_file_name: Name of the image that should be uploaded. Set to `None` to use the file name derived from `image_file` if provided as a path, otherwise a name is required.
         :param status: Status of the newly created item. Should be one of `['uploaded', 'annotated', 'reviewed', 'ignored']`. Set to `None` to automatically set to `reviewed` or `uploaded` in case the annotations are empty.
+        :param created_date: Creation date and time as datetime object or number (seconds, POSIX timestamp) of the dataset item (upload time is used if not provided).
         :return: ID of the newly created item or `-1` in case it could not be created.
         """
 
-        if image_file_name is None:
-            image_file_name = os.path.split(image_file_path)[-1]
+        if annotations is None:
+            annotations = {}
+        if isinstance(image_file, bytes):
+            if not isinstance(image_file_name, str):
+                raise ValueError('If an image is passed in bytes form, an "image_file_name" (str) has to be provided.')
+            files = {
+                'data': (image_file_name, image_file)
+            }
+        else:
+            if image_file_name is None:
+                image_file_name = os.path.split(image_file)[-1]
 
-        files = {
-            'data': (image_file_name, open(image_file_path, 'rb'))
-        }
+            files = {
+                'data': (image_file_name, open(image_file, 'rb'))
+            }
 
         if status is None:
             status = 'reviewed' if annotations else 'uploaded'
@@ -484,6 +548,18 @@ class DLDSClient:
             'status': status
         }
 
+        if created_date is not None:
+            if type(created_date) in (int, float):
+                created_date_dt = datetime.fromtimestamp(created_date)
+            else:
+                created_date_dt = created_date
+
+            if type(created_date_dt) != datetime:
+                raise ValueError('"created_date" must be a datetime object or a number')
+
+            created_date_str = created_date_dt.isoformat()
+            dataset_item['created_date'] = created_date_str
+
         try:
             response = self.http.post(self.server_url + '/dataset-items/', data=dataset_item, files=files)
             response.raise_for_status()
@@ -524,6 +600,147 @@ class DLDSClient:
 
         return item_details
 
+    def delete_dataset_item(self, item_id):
+        """Delete a dataset item.
+
+        :param item_id: id of the dataset item
+        :return: 'True' if success else 'False'
+        """
+
+        try:
+            response = self.http.delete(f'{self.server_url}/dataset-items/{item_id}')
+            response.raise_for_status()
+            return True
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return False
+
+    def download_dataset_item(self, item_id, image_dir: Optional[str] = None, annotation_dir: Optional[str] = None,
+                              accepted_status=None, return_item: bool = False):
+        """Download a dataset item into the specified directories (image and annotations).
+
+        :param item_id: ID of the dataset item to download.
+        :param image_dir: Output directory for the image.
+        :param annotation_dir: Output directory for the annotations.
+        :param accepted_status: List of accepted status. If the item status is not in this list, it will not be downloaded.
+        :return:
+        """
+        if image_dir is None and annotation_dir is None and not return_item:
+            raise ValueError('arguments not correct')
+
+        # load item details
+        try:
+            response = self.http.get(self.server_url + '/dataset-items/{}'.format(item_id))
+            response.raise_for_status()
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+            return None
+
+        item_details = json.loads(response.content)
+        if accepted_status is not None:
+            if item_details['status'] not in accepted_status:
+                return None
+        image_url = item_details['data']
+
+        # download image
+        try:
+            response = self.s3_http.get(image_url, stream=True)
+            response.raise_for_status()
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        item_id = item_details['id']
+        dataset_id = item_details['dataset_id']
+        item_name = item_details['name'].split('/')[-1]
+        item_name = '{}_{}_{}'.format(dataset_id, item_id, item_name)
+
+        image_bytes: Optional[bytes] = response.content
+        if image_bytes is None:
+            image_bytes = bytes()
+            for chunk in response.iter_content(4096):
+                image_bytes += chunk
+
+        if image_dir:
+            try:
+                os.makedirs(image_dir)
+            except FileExistsError:
+                pass
+            image_file_path = os.path.join(image_dir, item_name)
+
+            with open(image_file_path, 'wb') as f:
+                f.write(image_bytes)
+
+        if annotation_dir:
+            try:
+                os.makedirs(annotation_dir)
+            except FileExistsError:
+                pass
+            # write annotations to file
+            annotations = item_details['annotations']
+            annotation_file_name = os.path.splitext(item_name)[0] + '.json'
+            annotation_file_path = os.path.join(annotation_dir, annotation_file_name)
+            with open(annotation_file_path, 'w') as f:
+                json.dump(annotations, f, indent=2)
+
+        if return_item:
+            item_details['image'] = image_bytes
+            return item_details
+
+        return None
+
+    def add_dataset_items_to_subset(self, item_ids, subset_id):
+        """Add an existing dataset item to a subset.
+
+        :param item_ids: List of item IDs or a single item ID to be added
+        :param subset_id: ID of target subset
+        :return: 'True' if success, else 'False'
+        """
+        if not isinstance(item_ids, list):
+            if isinstance(item_ids, int):
+                item_ids = [item_ids]
+            else:
+                raise ValueError('Item IDs has to be a list of integers or an integer')
+
+        items = {
+            'items': item_ids
+        }
+
+        try:
+            response = self.http.post(f'{self.server_url}/data-subsets/{subset_id}/items/', json=items)
+            response.raise_for_status()
+            return True
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return False
+
+    def remove_dataset_items_from_subset(self, item_ids, subset_id):
+        """Remove a dataset item from a subset.
+
+        :param item_ids: List of item IDs or a single item ID to be removed
+        :param subset_id: ID of target subset
+        :return: 'True' if success, else 'False'
+        """
+        if not isinstance(item_ids, list):
+            if isinstance(item_ids, int):
+                item_ids = [item_ids]
+            else:
+                raise ValueError('Item IDs has to be a list of integers or an integer')
+
+        items = {
+            'items': item_ids
+        }
+
+        try:
+            response = self.http.post(f'{self.server_url}/data-subsets/{subset_id}/remove_items/', json=items)
+            response.raise_for_status()
+            return True
+        except rq.exceptions.RequestException as e:
+            logger.error(e)
+
+        return False
+
     def get_model(self, model_id):
         """Get information about a model.
 
@@ -1567,7 +1784,7 @@ class DLDSClient:
         :return: Status code. 0: import completed, -1: error
 
         [1] Lin et al., "Microsoft COCO: Common Objects in Context", 2014.
-            http://cocodataset.org/
+            https://cocodataset.org/
         """
 
         logger.info('Import COCO data')
@@ -1647,8 +1864,8 @@ class DLDSClient:
         objects = []
         if image_id not in coco_annotation_ids:
             return
-        for i, id in enumerate(coco_annotation_ids[image_id]):
-            a = coco_annotations[id]
+        for i, cid in enumerate(coco_annotation_ids[image_id]):
+            a = coco_annotations[cid]
 
             coco_class_id = a['category_id']
             class_label_name = coco_class_id_map[coco_class_id]
