@@ -397,7 +397,7 @@ class DLDSWorker:
                 model_env['DLDS_JOB_ID'] = str(self.current_job.get('id'))
                 model_env['DLDS_SERVER_URL'] = server_url
                 model_env['DLDS_DATASET_DIR'] = dataset_dir
-                model_env['DLDS_MODEL_CLASS'] = 'dldstraining.TrainerModel'
+                model_env['DLDS_MODEL_CLASS'] = os.environ.get('DLDS_MODEL_CLASS', 'dldstraining.TrainerModel')
 
                 username = self.config.get('username')
                 password = self.config.get('password')
